(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 12.2' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[     77877,       1518]
NotebookOptionsPosition[     76510,       1486]
NotebookOutlinePosition[     77008,       1504]
CellTagsIndexPosition[     76965,       1501]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[BoxData[{
 RowBox[{
  RowBox[{"PlanckConstant", "=", 
   RowBox[{"6.62606957", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "34"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"PlanckConstantReduced", "=", 
   FractionBox["PlanckConstant", 
    RowBox[{"2", "  ", "\[Pi]"}]]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"FineStructureConstant", "=", 
   RowBox[{"7.2973525698", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "3"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"ElectronMass", "=", 
   RowBox[{"9.10938291", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "31"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"GravitationalConstant", "=", 
   RowBox[{"6.67384", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "11"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"ProtonMass", "=", 
   RowBox[{"1.672621777", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "27"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"AtomicMassUnit", "=", 
   RowBox[{"1.66053878283", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "27"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"BohrRadius", "=", 
   RowBox[{"0.52917721092", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "10"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"BohrMagneton", "=", 
   RowBox[{"927.400968", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "26"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"BoltzmannConstant", "=", 
   RowBox[{"1.3806488", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "23"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"StandardGravityAcceleration", "=", "9.80665"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"SpeedOfLight", "=", "299792458"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"StefanBoltzmannConstant", "=", 
   RowBox[{"5.670373", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "8"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"ElectronCharge", "=", 
   RowBox[{"1.602176565", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "19"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"VacuumPermeability", "=", 
   RowBox[{"4", "\[Pi]", " ", 
    SuperscriptBox["10", 
     RowBox[{"-", "7"}]]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"DielectricConstant", "=", 
   FractionBox["1", 
    RowBox[{
     SuperscriptBox["SpeedOfLight", "2"], " ", "VacuumPermeability"}]]}], 
  ";"}], "\n", 
 RowBox[{
  RowBox[{"ElectronGyromagneticFactor", "=", 
   RowBox[{"-", "2.00231930436153"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"AvogadroConstant", "=", 
   RowBox[{"6.02214129", " ", 
    SuperscriptBox["10", "23"]}]}], ";"}]}], "Input",
 CellChangeTimes->{
  3.9462898535670986`*^9, {3.946289960548875*^9, 3.9462899947154675`*^9}, {
   3.9462971681147823`*^9, 3.9462971786756124`*^9}, {3.94629883673812*^9, 
   3.9462988374090405`*^9}},
 CellLabel->"In[3]:=",ExpressionUUID->"62f7b3c0-151e-42a7-bd14-2c7024749c6f"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"scatteringLength", "[", "B_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"abkg", ",", "resonanceB", ",", "resonancewB", ",", "as"}], 
     "}"}], ",", 
    RowBox[{"(*", 
     RowBox[{"Set", " ", "background", " ", "scattering", " ", "length"}], 
     "*)"}], 
    RowBox[{
     RowBox[{"abkg", "=", 
      RowBox[{"85.5", "*", "BohrRadius"}]}], ";", 
     RowBox[{"(*", 
      RowBox[{
      "BohrRadius", " ", "should", " ", "be", " ", "defined", " ", 
       "beforehand"}], "*)"}], 
     RowBox[{"(*", 
      RowBox[{"Resonance", " ", "positions", " ", "and", " ", "widths"}], 
      "*)"}], 
     RowBox[{"resonanceB", "=", 
      RowBox[{"{", 
       RowBox[{
       "1.295", ",", "1.306", ",", "2.174", ",", "2.336", ",", "2.591", ",", 
        "2.74", ",", "2.803", ",", "2.78", ",", "3.357", ",", "4.949", ",", 
        "5.083", ",", "7.172", ",", "7.204", ",", "7.134", ",", "76.9"}], 
       "}"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"resonancewB", "=", 
      RowBox[{"{", 
       RowBox[{
       "0.009", ",", "0.010", ",", "0.0005", ",", "0.0005", ",", "0.001", ",",
         "0.0005", ",", "0.021", ",", "0.015", ",", "0.043", ",", "0.0005", 
        ",", "0.130", ",", "0.024", ",", "0.0005", ",", "0.036", ",", "3.1"}],
        "}"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"(*", 
      RowBox[{"Get", " ", "scattering", " ", "length"}], "*)"}], 
     RowBox[{"as", "=", 
      RowBox[{"abkg", "*", 
       RowBox[{"Product", "[", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{
            RowBox[{"resonancewB", "[", 
             RowBox[{"[", "j", "]"}], "]"}], "/", 
            RowBox[{"(", 
             RowBox[{"B", "-", 
              RowBox[{"resonanceB", "[", 
               RowBox[{"[", "j", "]"}], "]"}]}], ")"}]}]}], ")"}], ",", 
         RowBox[{"{", 
          RowBox[{"j", ",", "1", ",", 
           RowBox[{"Length", "[", "resonanceB", "]"}]}], "}"}]}], "]"}]}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"(*", 
      RowBox[{"Return", " ", "scattering", " ", "length"}], "*)"}], 
     RowBox[{"as", "/", "BohrRadius"}]}]}], "]"}]}], "\n", 
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"scatteringLength", "[", "b", "]"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{"b", ",", " ", "0", ",", " ", "8"}], "}"}], ",", " ", 
   RowBox[{"PlotRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", "2.5"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "150"}], "}"}]}], "}"}]}], ",", 
   RowBox[{"FrameLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
     "\"\<B (G)\>\"", ",", "\"\<Scattering Length (x BohrRadius)\>\""}], 
     "}"}]}], ",", 
   RowBox[{"PlotRangeClipping", "\[Rule]", "True"}], ",", 
   RowBox[{"LabelStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"FontSize", "\[Rule]", "14"}], "}"}]}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", " ", 
   RowBox[{"PlotTheme", "\[Rule]", "\"\<Classic\>\""}], ",", 
   RowBox[{"TicksStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"FontSize", "\[Rule]", "14"}], "}"}]}], ",", 
   RowBox[{"PlotStyle", "\[Rule]", 
    RowBox[{"{", "Thick", "}"}]}], ",", 
   RowBox[{"GridLines", "\[Rule]", "Automatic"}], ",", 
   RowBox[{"ImageSize", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"800", ",", "495"}], "}"}]}], ",", 
   RowBox[{"AspectRatio", "\[Rule]", "Full"}]}], "]"}]}], "Input",
 CellChangeTimes->{{3.946298626684884*^9, 3.946298627172211*^9}, {
  3.9462986848238554`*^9, 3.9462987262215877`*^9}, {3.946298822465828*^9, 
  3.9462988704810686`*^9}, {3.946298989364349*^9, 3.946298992368571*^9}, {
  3.9462990341129494`*^9, 3.946299053088297*^9}, {3.9462991029805555`*^9, 
  3.946299104687995*^9}, {3.946299143481086*^9, 3.9462993374952354`*^9}, {
  3.9462994016976433`*^9, 3.94629940724636*^9}, {3.9462994625206733`*^9, 
  3.946299472231315*^9}, {3.9462995502251225`*^9, 3.946299611028179*^9}, {
  3.946299685440056*^9, 3.946299719781707*^9}, {3.946299802805818*^9, 
  3.946299822797445*^9}},
 CellLabel->"In[21]:=",ExpressionUUID->"461e8a7b-9e39-408e-a5f3-38960ab44ff7"],

Cell[BoxData[
 GraphicsBox[{{{}, {}, 
    TagBox[
     {RGBColor[0.24720000000000014`, 0.24, 0.6], Thickness[Large], Opacity[
      1.], LineBox[CompressedData["
1:eJwVkgk01fkbxmmVJLsYiRb9m3QtjTV6X25RlhJTvr+LGxcjLUO2UEgJhWnG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       "]], LineBox[CompressedData["
1:eJwtkQs0VQkXx8+hB3mVxEWIhKb08Ka0N3lF8rz3nENKRqIiUZLy1jRJEymi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       "]], LineBox[CompressedData["
1:eJwVjH1MzHEAh/t+J0tcIzqytI7IxtSWbkr4yHlLOM7LfX+9TVM5L5cjyejF
eUmSjju6smxc6UTSWq68FL0petEqL6n0si7KS+U2cWnyx7Pnn2ePICRieyi1
sLBwG+e/h5taNGoZAcQJQ4NRO+Db+r0+SUEQ06fZ3RO7G1/uKWM1UQT5ygaz
lifFhRj+koyTBD5ehtL+YClctua0Z8UTFBl+RM68LkWV08rkB+cIBC8KrXqe
ShE23OhtuEhwtKJGHNQqhWVF6NfSFAK/dR/La4akyLz2+0a1huBc0efWyZYM
ovDkTY1pBBJXuUlix9CzTGD+cJOg6bHIUTaPQWldmNOtI1hsTQ80uTII2jZw
A3qCZlW87U8vhue5bZNMuQR13ZZ2NmsZxsQTZJZFBGZjT+SfXQwZc7WzbJ4R
7DIYQ2yDGbxNi6r5ZQS1G56XFIUxnEyVuCysJch7GD3FHMkw0n6717eTQGw8
W/bkMkNqnvCapJcgvnijpE7DIDz9ShTQT7DiRGNgVTrDMedh3SETwY7EMNWZ
OwzD+1eHqKwoVs25ODHzGcMV75ZpaTyKGSPitohyBjcb2YtbthSKtdEKUQ2D
PF/lVOBAkWlccpXfwjDw61NHixuFe65W9eYbw6PwlDJfIYVX79HX3SaG+Hcr
75Qup8g6XFq5aJSBb8g4dHcdxbcTH7ITJ3PoWrB5m6MfxTFZoZNhOof710eX
qrdR3DsVdGuzAwefKP/RUwEUnn0vE6xdOfCM1p1DeyiipLpEmSeH9zsfl4eF
U3iEfLbPFnHQVcqyPx6kaCgOx28xB7mHfZL4CAVP4tOpD+LgmVUtrzw+3r+d
3j5bzmGCXfR2rxiK0Dh/wac4Dg1nXYR5SgpViez4Kw2HdNNbe+cLFOu7WNL8
BxxC957/q02m0A1sibWq5+DW7NHFU1MMTl2onTPCwbymt0KppRjqd+5QuPuj
qkCj/5VBEZfAFyal+OPKPNGlgzqKaen6jenCAASof0Z06SnkEWMOHY2BmNuq
8OvLGf/vU4+MrQ7GP6x6WKE=
       "]], LineBox[CompressedData["
1:eJwVjntQVGUAR5f7fUYtRigzIk8VBFITsUC0kP3xGkFjVeR1792L8k5nB5Qo
RCUfgAMGrLmMkokgUMrmCjiCwDrIaw2XAVcyBpsMho0WNg1wBIkRSc/MmfPv
WRWfFp7EiEQivze+bfUhrS/XwGB57MF3ZsWRuDNiZo9mBnKHSzlSSTROh9e/
vtXKYJM00KgWYrC7Y9/Imi4GdZ5r5UMjMXD42Ore5fsMnicGkG0bWRiv3K2x
fsDgLrMjZDSJxc0laUX5jxgcTkjQRBeyyD7pdGj+MYPdPV/fWq9iETLVG5E+
xCC+y9kjuYOFdVz2ZuNfDM4taGw6B1j8qf/IQWZi0Gjx+7AwxqIGfyzoJxgE
NSmvTbxkkVH3rSF4mkF2ymLJy0UcJCs/+6VljkG/l0dW31IO4rMm1QYRwck1
zT2NThzK00LTbS0I9CmyonQvDgeGZiOLrQjCQ0Pd8v04eO28toUsI7DsE9Vv
DuGg8zA3e7aSoFm3X1rHcZh51qVo20RQm/XKNSSbQ7uQ8aW3L8G+u0rfyjwO
hb0u0Sp/gtYb0XamIg7O6lNOJZ8ThFzUJUVe5hAmhzolnqB/fOz6QiuHapNG
Z1lMMKwTTv9jxiP1wk6XA0qCbsWvoaliHj5BhqPaUgJJQ88Rc2seujLx+mNV
BJ6Fhqqc1TymdrGK8SaCsJ+qgh+G8GiZf2oMaiWQNrpovcN55KpOoKKT4HgG
G97K81hOr01F9RGcinaJC0zj4Xd7Zk+XgSBPo05nS3m8m3jm+opxgjiVaaVd
FY9+K6dFR/8lsDJW5Cy5wSPpQHDjxv8ILrXrTXVdPAodS5aXW1KMGm4obKd5
ROncD81ZU7S9943rjJkMKzI19yNtKdY9GFiz7gMZbupHjixeTVHmP3qvaa0M
j3M2PsnaQmHeG3A1LEWGSk+t929+FOIdY1UDmTLIn8QUewZRfHq7clBcIIPI
54TEKKVwlsc0dKpl0BmsSwMiKAJfjLFZ7TIoz16dLGMpnhdptvkNyOBm6rsS
kUixdau9axkRMHk+fq52P8W5Rq3/jw4CWgJnwi3SKH6IHcur8BGQO1nwc3IG
hdVDheJFhICwMkfakUUxwecvuH8lYNn2epnj8Te/11O5qVIBwzNBDYdzKV43
ct3SdgGqqsH3HxVQ5Pe6FKknBGTskidvUFBYXDAMebjFwm9+ofVMCUWojVLi
/kUszFVKm7+/p/gu87zeXhMLfZT7Qf9yCsng9NKnH+7FRaLpvlRNcay+o2y9
di8S66SrZmsoEhbsPHNj98FDGMnaU0vxqmTozuXtcQi7/Yk68iaF6C1t8fgf
YlKrLQ==
       "]], LineBox[CompressedData["
1:eJwVj30s1HEAhy9+35Vm5vVICDVOZ7cV/zhmnyhZOxt68XYvv995aazNy5i3
pJCRyoomtqhYZhFuixJHXPJ27tKLvOStI8xLdZcyvfFsz56/HwdpfHC0HovF
ct52p43+sQ1lnRQcwhNHq6eCcMb4yGSXgsJqp84lqDYEdh0Kd5sBCgqDvNge
WRgWY0NvpKoorM3Hj/6Wh0PGXpkfeUfBs0k+IzofgcyebG/eOIULTBm7ZioC
xxPMygqnKRTUW60G8oUwsq1d12gomDQpG82uCoFzuZ8+L1HoGDWyUXUJkVQs
GZxbo/DN1WXf0g8havo8n89qKbQV22pOOYnwYZdV7cwvCh4tPlu+p0XY46kr
nf5DYUwy4daTJQI/WZ0zpUfwpSGMn/ZIhMqFAvGkIcFc6NCVDJ0I6gPRggkT
AjU3mJ1jLYZe2DH+OJugLqa1ygdixAxuWny0J9hMPPq9v0gMXlOc8q07QXJ/
tbbCWQJ6ye/FiAfB8DqXKwqU4Lbjwbo33gRe+pr7VhkSbJRO5Kn8CUoEG/qb
agk4w61JwwEEr3ObLcR/JQjfXUorgwluLQ/1VrJpyNMFXoNCAtN2m6Q/J2l8
lXEODzAELMtDW6ZSGo4rlFV/DEFRnyDKLYtGvrhD25tAkN48ptC20HhWVj77
KoXAwlvR2/eexrI6RaXIIDC8Jss126AR4Mt73J1HsN90SHfHk8Gli3vLXxYS
XHfV2dsxDJqeLuR33SQ4a+5UIi9gMLfWndxZQvCAF2D6RMbAnFMlld8lWJyM
TOVOM/BjMgM77hFYJvy8nG0kRVpFiHf7Q4LIqFmvf5Bi0tDYuq2eoOKEQSyn
WQqj2TjlYsP2zw5dUvwHh1IY7A==
       "]], LineBox[CompressedData["
1:eJwtlXlQkwcTxiPS980bKaDlA1pCOVQah0OqXIrIw1kDFElIQkgIJCCMVMsl
akGUs8GjyKGW8kkpRUBEioQGkEIpCgJyCH6KGZRaFJQKQkGRqlBtOv125pn9
Y2f39+z+s2bhsdxIDRqNFqXWPznXdPWdskICtP/HN9MruMNnCZxwLPtJo9QQ
i1df7cP3BOa/ZO88/uMG8P47f+ZiJQGjUA7Dq3AT6uOfNOnXEKgwOl/ue8YR
OuwHI+kKAlVt20yaXjljr+nI0tNGAl1hPE7emCt6/xwyFrYSGK3eb1twyQ2s
wR7XjisEHDJLbr5x94C8sl1m000gwz4kyVDhiYnDlzOL+gmEKV9Gm771ghu/
rkLzfwRkNoUI/fgTlFhVdceqCDzbmf2n0HcHlleWPrk7SsBwraKEFsCG6F7h
Ku+HBMp9ZxMG3HxwuT7XWjFJQI+jilaZ+0L/ePZO5gwBccTQjt8XfLFPlhqf
/YxAwBGL5mPNfrDRjW2Q/EWgz5lVUW7ij68mo1Q9GiQOSIW/sTv8MdUW+moz
ncQLF2p5XLITFZ/7uzDeI9HXsVae4RIADS/vsP2GJJg/WJfEpAZAytyePmZM
grMmKHBDWwCYfdbXGlkkggLn92ls5iC5bP2kmQ0J+eOg0znRHKiSjKmczSQ6
jcocr37LQQHr3U8jtpOIbov0GXjLwdxbzdgbHiSq4y1sZ6y58Fct521hkzBX
ynk+Ii4o+dPbOjwSeZ0m+htruYiSTCwmB6v5RfWeU8NcdNqNGj4OJdEwcMrP
Z5mLtPG+kNZoEhbXs7x8vQKxiIsTu1NJhCStO4jeQLSWT589mEViqpflYDAZ
iAy6VaD8GIkWbcs+25U8aA/VXDl3isRfTim23C08WITVltw/T8L6ebGOUwkP
T6/O8mdqSKQsbbXXa+Sh3mLju8sKEiO3ph/ED/CwffbSofdbSSgXG3Ouv+ZB
cFgh5A2RoNprV6cF8MF8MK8TMUxieNdcBy2Sj4eem7rj75Kg2634wjKJjxit
H+1zJ0hURBuPnyzlQ35Wuab3JYkj5smShWk+Gpqb+l3N6Fj7uLZq+aAAycYv
M/0t6CiQD87UHRUA6U7OEks6EnVaB1yLBOjf0Xwh2Z6OP762dF1sFmBC9ZO8
gU3HglXaNO21AHqLP8MygY6pMpskHAhC4qZOpX6Huj8mLVa5R4goO9PS/B41
b3/eM0WKEEKHlBNaN+i4YZMz0p0jhPNWu3CNEToKPzIReNcJoeFerjM7S8et
3w2ybi0KseBBW9q9QMfG20VhI/RgPPIKeTz+Su3nedx6LaNg9LD1fh7RpNDl
afwfTQQjl5P12bUPKOzKUH1z+qtgpAWO8WFKYW56TaCwNBgJ/G1uLespeDt0
6UUogyEIXjBQ2FLIV1W5Sn4NBlO261qxN4WLKxQrxZtF0I5orzPwo/Bh48k7
/WwRaJHM4gIOhbFK+uAFqQjju28nHA2h4K67vaYlV4TqOE/TxAR1/WL5/Mxz
EYoTSlf9cZDCxqpJ8yhtMXISlxejD1PYtkvOT90gRtwXyoGwoxRKhvtqmmRi
hCfrNt/NocCXMKueHBGDl7K3nH+KgtHJVZ89LBbDMW3dId8SCpHHbzH498Rg
ZaRFdZ1T7xfQ5ji5JMYHWaMctwsUaooeTsUZh+BN9mmWo1LNr9RKrYgIQWeu
5h3zXgpHr07YxbwJQWO+9Mq3gxR8mvOH81kSVJ1qrTEcpmASkDfrwZPgRGFi
pvaY+p7MM1a/1EmQUjQUc+wRBYGTrUHnAwlizlqJNKcpuKia8gb1QsH9bsL2
9QsKDo4HXAvSQ+H5PZj7lyhk5VX+UtISCvtzxeQcjYHSIbdAyctQGJ7n3X+0
ioHurVnzN5LDwLhQd126mgFccl1PuxKG5Wqthnv6DMgOJTTGvSPFb7WdJ26a
MbDFycw7ky/F0Izm000fMZA5epzukCZFu5WX32kr9Xwiw3jvD1LU7fmy5sXH
DIjujvtE/CpFafU1rSBHBu5rEwkyXRnynrzz+eVtDEToOs81eMuQzvIeeN+d
Acqnnt2SJkN4ZVfOqB8DVh2N9OwV4eA+ImZduAzk7dkz788Oh8e6T/y/C2KA
1d9TefPrcGQsNXa4ixj//pP2cPwNZf+2KQ==
       "]], LineBox[CompressedData["
1:eJwtVmk0lQsXJlS85xzO++YqjUijWyqFkp4i83CcwZkcx3FDKSRDhoqUJHJv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       "]], LineBox[CompressedData["
1:eJwtVnk0lQv3pgwlKZz35Zz3lAwXZUhKQpcnJEM453DOMR5jVNccbjJliitC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       "]], LineBox[CompressedData["
1:eJwVk3s4lfkWx91ymyK3GXXe3/uelKQkGQyZ9I3dtm3bvryvaXehvV1Kj3HJ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       "]], 
      LineBox[{{4.951551020408163, 139.37596259923768`}, {4.952725024990185, 
       150.}}], 
      LineBox[{{7.196012175022929, 0.}, {7.19644978787316, 
       0.680122364067862}, {7.198928048521438, 4.3576005303963}, {
       7.201406309169715, 8.366240097856897}, {7.201448979591836, 
       8.448952703871816}}], 
      LineBox[{{2.794950454536869, 0.}, {2.7969872680162746`, 
       52.52496966501432}, {2.7991671838755243`, 150.}}]},
     Annotation[#, "Charting`Private`Tag$11333#1"]& ], {}}, {}},
  AspectRatio->Full,
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{0, 0},
  AxesStyle->Directive[
    GrayLevel[0], 
    AbsoluteThickness[0.2]],
  BaseStyle->Automatic,
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox["\"Scattering Length (x BohrRadius)\"", TraditionalForm], None}, {
     FormBox["\"B (G)\"", TraditionalForm], None}},
  FrameStyle->Directive[
    GrayLevel[0], 
    AbsoluteThickness[0.2]],
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  FrameTicksStyle->GrayLevel[0],
  GridLines->{Automatic, Automatic},
  GridLinesStyle->Automatic,
  ImagePadding->All,
  ImageSize->{800, 495},
  LabelStyle->{FontSize -> 14},
  Method->{
   "DefaultBoundaryStyle" -> Automatic, 
    "DefaultGraphicsInteraction" -> {
     "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
      "Effects" -> {
       "Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2}, 
        "Droplines" -> {
         "freeformCursorMode" -> True, 
          "placement" -> {"x" -> "All", "y" -> "None"}}}}, "DefaultMeshStyle" -> 
    AbsolutePointSize[6], "PointSizeFunction" -> None, "ScalingFunctions" -> 
    None, "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Identity[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Identity[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{0, 2.5}, {0, 150}},
  PlotRangeClipping->True,
  PlotRangePadding->{{0, 0}, {0, 0}},
  Ticks->{Automatic, Automatic},
  TicksStyle->{FontSize -> 14}]], "Output",
 CellChangeTimes->{
  3.9462995040442867`*^9, {3.946299583927579*^9, 3.946299611442181*^9}, 
   3.9462996528493867`*^9, {3.9462996867124157`*^9, 3.946299720405142*^9}, {
   3.94629980609776*^9, 3.946299830136408*^9}, 3.9463000847462177`*^9, 
   3.946301602975066*^9, {3.94654225425524*^9, 3.9465422604126263`*^9}},
 CellLabel->"Out[22]=",ExpressionUUID->"fe3d2ab1-2e60-4c87-b304-9af09b512293"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"scatteringLength", "[", "b", "]"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{"b", ",", " ", "0", ",", " ", "8"}], "}"}], ",", " ", 
   RowBox[{"PlotRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1.317", ",", "2.173"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "150"}], "}"}]}], "}"}]}], ",", 
   RowBox[{"FrameLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
     "\"\<B (G)\>\"", ",", "\"\<Scattering Length (x BohrRadius)\>\""}], 
     "}"}]}], ",", 
   RowBox[{"PlotRangeClipping", "\[Rule]", "True"}], ",", 
   RowBox[{"LabelStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"FontSize", "\[Rule]", "14"}], "}"}]}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", " ", 
   RowBox[{"PlotTheme", "\[Rule]", "\"\<Classic\>\""}], ",", 
   RowBox[{"TicksStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"FontSize", "\[Rule]", "14"}], "}"}]}], ",", 
   RowBox[{"PlotStyle", "\[Rule]", 
    RowBox[{"{", "Thick", "}"}]}], ",", 
   RowBox[{"GridLines", "\[Rule]", "Automatic"}], ",", 
   RowBox[{"ImageSize", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"800", ",", "495"}], "}"}]}], ",", 
   RowBox[{"AspectRatio", "\[Rule]", "Full"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.946300123196314*^9, 3.9463001547081313`*^9}, {
  3.946300433237008*^9, 3.9463004333250217`*^9}, {3.946300466694854*^9, 
  3.9463004847154093`*^9}, {3.946300534787926*^9, 3.9463005348354373`*^9}},
 CellLabel->"In[23]:=",ExpressionUUID->"e3efdc46-0d6e-4d16-ae45-a127d2ef7398"],

Cell[BoxData[
 GraphicsBox[{{{}, {}, 
    TagBox[
     {RGBColor[0.24720000000000014`, 0.24, 0.6], Thickness[Large], Opacity[
      1.], LineBox[CompressedData["
1:eJwVkgk01fkbxmmVJLsYiRb9m3QtjTV6X25RlhJTvr+LGxcjLUO2UEgJhWnG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       "]], LineBox[CompressedData["
1:eJwtkQs0VQkXx8+hB3mVxEWIhKb08Ka0N3lF8rz3nENKRqIiUZLy1jRJEymi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       "]], LineBox[CompressedData["
1:eJwVjH1MzHEAh/t+J0tcIzqytI7IxtSWbkr4yHlLOM7LfX+9TVM5L5cjyejF
eUmSjju6smxc6UTSWq68FL0petEqL6n0si7KS+U2cWnyx7Pnn2ePICRieyi1
sLBwG+e/h5taNGoZAcQJQ4NRO+Db+r0+SUEQ06fZ3RO7G1/uKWM1UQT5ygaz
lifFhRj+koyTBD5ehtL+YClctua0Z8UTFBl+RM68LkWV08rkB+cIBC8KrXqe
ShE23OhtuEhwtKJGHNQqhWVF6NfSFAK/dR/La4akyLz2+0a1huBc0efWyZYM
ovDkTY1pBBJXuUlix9CzTGD+cJOg6bHIUTaPQWldmNOtI1hsTQ80uTII2jZw
A3qCZlW87U8vhue5bZNMuQR13ZZ2NmsZxsQTZJZFBGZjT+SfXQwZc7WzbJ4R
7DIYQ2yDGbxNi6r5ZQS1G56XFIUxnEyVuCysJch7GD3FHMkw0n6717eTQGw8
W/bkMkNqnvCapJcgvnijpE7DIDz9ShTQT7DiRGNgVTrDMedh3SETwY7EMNWZ
OwzD+1eHqKwoVs25ODHzGcMV75ZpaTyKGSPitohyBjcb2YtbthSKtdEKUQ2D
PF/lVOBAkWlccpXfwjDw61NHixuFe65W9eYbw6PwlDJfIYVX79HX3SaG+Hcr
75Qup8g6XFq5aJSBb8g4dHcdxbcTH7ITJ3PoWrB5m6MfxTFZoZNhOof710eX
qrdR3DsVdGuzAwefKP/RUwEUnn0vE6xdOfCM1p1DeyiipLpEmSeH9zsfl4eF
U3iEfLbPFnHQVcqyPx6kaCgOx28xB7mHfZL4CAVP4tOpD+LgmVUtrzw+3r+d
3j5bzmGCXfR2rxiK0Dh/wac4Dg1nXYR5SgpViez4Kw2HdNNbe+cLFOu7WNL8
BxxC957/q02m0A1sibWq5+DW7NHFU1MMTl2onTPCwbymt0KppRjqd+5QuPuj
qkCj/5VBEZfAFyal+OPKPNGlgzqKaen6jenCAASof0Z06SnkEWMOHY2BmNuq
8OvLGf/vU4+MrQ7GP6x6WKE=
       "]], LineBox[CompressedData["
1:eJwVjntQVGUAR5f7fUYtRigzIk8VBFITsUC0kP3xGkFjVeR1792L8k5nB5Qo
RCUfgAMGrLmMkokgUMrmCjiCwDrIaw2XAVcyBpsMho0WNg1wBIkRSc/MmfPv
WRWfFp7EiEQivze+bfUhrS/XwGB57MF3ZsWRuDNiZo9mBnKHSzlSSTROh9e/
vtXKYJM00KgWYrC7Y9/Imi4GdZ5r5UMjMXD42Ore5fsMnicGkG0bWRiv3K2x
fsDgLrMjZDSJxc0laUX5jxgcTkjQRBeyyD7pdGj+MYPdPV/fWq9iETLVG5E+
xCC+y9kjuYOFdVz2ZuNfDM4taGw6B1j8qf/IQWZi0Gjx+7AwxqIGfyzoJxgE
NSmvTbxkkVH3rSF4mkF2ymLJy0UcJCs/+6VljkG/l0dW31IO4rMm1QYRwck1
zT2NThzK00LTbS0I9CmyonQvDgeGZiOLrQjCQ0Pd8v04eO28toUsI7DsE9Vv
DuGg8zA3e7aSoFm3X1rHcZh51qVo20RQm/XKNSSbQ7uQ8aW3L8G+u0rfyjwO
hb0u0Sp/gtYb0XamIg7O6lNOJZ8ThFzUJUVe5hAmhzolnqB/fOz6QiuHapNG
Z1lMMKwTTv9jxiP1wk6XA0qCbsWvoaliHj5BhqPaUgJJQ88Rc2seujLx+mNV
BJ6Fhqqc1TymdrGK8SaCsJ+qgh+G8GiZf2oMaiWQNrpovcN55KpOoKKT4HgG
G97K81hOr01F9RGcinaJC0zj4Xd7Zk+XgSBPo05nS3m8m3jm+opxgjiVaaVd
FY9+K6dFR/8lsDJW5Cy5wSPpQHDjxv8ILrXrTXVdPAodS5aXW1KMGm4obKd5
ROncD81ZU7S9943rjJkMKzI19yNtKdY9GFiz7gMZbupHjixeTVHmP3qvaa0M
j3M2PsnaQmHeG3A1LEWGSk+t929+FOIdY1UDmTLIn8QUewZRfHq7clBcIIPI
54TEKKVwlsc0dKpl0BmsSwMiKAJfjLFZ7TIoz16dLGMpnhdptvkNyOBm6rsS
kUixdau9axkRMHk+fq52P8W5Rq3/jw4CWgJnwi3SKH6IHcur8BGQO1nwc3IG
hdVDheJFhICwMkfakUUxwecvuH8lYNn2epnj8Te/11O5qVIBwzNBDYdzKV43
ct3SdgGqqsH3HxVQ5Pe6FKknBGTskidvUFBYXDAMebjFwm9+ofVMCUWojVLi
/kUszFVKm7+/p/gu87zeXhMLfZT7Qf9yCsng9NKnH+7FRaLpvlRNcay+o2y9
di8S66SrZmsoEhbsPHNj98FDGMnaU0vxqmTozuXtcQi7/Yk68iaF6C1t8fgf
YlKrLQ==
       "]], LineBox[CompressedData["
1:eJwVj30s1HEAhy9+35Vm5vVICDVOZ7cV/zhmnyhZOxt68XYvv995aazNy5i3
pJCRyoomtqhYZhFuixJHXPJ27tKLvOStI8xLdZcyvfFsz56/HwdpfHC0HovF
ct52p43+sQ1lnRQcwhNHq6eCcMb4yGSXgsJqp84lqDYEdh0Kd5sBCgqDvNge
WRgWY0NvpKoorM3Hj/6Wh0PGXpkfeUfBs0k+IzofgcyebG/eOIULTBm7ZioC
xxPMygqnKRTUW60G8oUwsq1d12gomDQpG82uCoFzuZ8+L1HoGDWyUXUJkVQs
GZxbo/DN1WXf0g8havo8n89qKbQV22pOOYnwYZdV7cwvCh4tPlu+p0XY46kr
nf5DYUwy4daTJQI/WZ0zpUfwpSGMn/ZIhMqFAvGkIcFc6NCVDJ0I6gPRggkT
AjU3mJ1jLYZe2DH+OJugLqa1ygdixAxuWny0J9hMPPq9v0gMXlOc8q07QXJ/
tbbCWQJ6ye/FiAfB8DqXKwqU4Lbjwbo33gRe+pr7VhkSbJRO5Kn8CUoEG/qb
agk4w61JwwEEr3ObLcR/JQjfXUorgwluLQ/1VrJpyNMFXoNCAtN2m6Q/J2l8
lXEODzAELMtDW6ZSGo4rlFV/DEFRnyDKLYtGvrhD25tAkN48ptC20HhWVj77
KoXAwlvR2/eexrI6RaXIIDC8Jss126AR4Mt73J1HsN90SHfHk8Gli3vLXxYS
XHfV2dsxDJqeLuR33SQ4a+5UIi9gMLfWndxZQvCAF2D6RMbAnFMlld8lWJyM
TOVOM/BjMgM77hFYJvy8nG0kRVpFiHf7Q4LIqFmvf5Bi0tDYuq2eoOKEQSyn
WQqj2TjlYsP2zw5dUvwHh1IY7A==
       "]], LineBox[CompressedData["
1:eJwtlXlQkwcTxiPS980bKaDlA1pCOVQah0OqXIrIw1kDFElIQkgIJCCMVMsl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       "]], LineBox[CompressedData["
1:eJwtVmk0lQsXJlS85xzO++YqjUijWyqFkp4i83CcwZkcx3FDKSRDhoqUJHJv
xJWURrrRJUqiDGWKlEQyJdV1FUlFGtDnW9+319prr2c9w9r739Zy92N7TJGT
k8uc7P/OSMOHSlhDQO7/VW10osjQgMBJ65WGr/4c2lS498nMU8YEklTUqLdp
M5FZyAz9bkqgv21aQIWfLv4ccWoTmxOYttdzvO3SrzhqkGh815qApU36gZyY
1QgJaDo1z5GAl23C/cJ/1mJ7Pvk9gkugOSv5ss5XIwg/skU9QgJBAV7MpNoN
sNFPKt4iJWCVkHqBkppivd9TzUvbCCxJu5hkJAOW/U2FK3kTyJ7mVGC0bzM0
33M6PP0IpI32t8gf2oIf3s2nl4ZN7tOTK9F2MQd3neTEiggCCXnTW8Z1tiJH
vvfImsMErA/p8oxfboVr6qj/xgQC5zwy6UXGlrj1W6TXliQCfCLj69XnllBb
qSyxTJ3M+86P3+ZnhYoqTSunCwQKsmsEWUesoXni4kbnLAK1j+aLLKfbYI+L
3hpxzqS+7x63+5ANFn7eOM+jkMBjpd64s9622F9aTXmXEDCbd6X64DNbPIt1
VPYrJ2ByrFu5daMdYhbIhkPqCFiYM7aP/bTDy/63b/c/JmBUOa/TUWwPo1t7
uqOaCbgp7sXcAnu8sz9cF/+CwFCZatEXgQPMZtHLT7yevFdlFkeW5YD0N8k3
U/oI7Krruqrw2QH24Vnnzn2a5LVUko5GOSLTQj/l8iiB4fR/e/OqHTHBvB13
dYwA545yuIMiC3lX6oJvTKXh6Wmv4FlmLFDP3tvVatIg6FtFP3aThV3ng7c0
zKeh+L5KM9XIQuWun4ZNC2nwOZNV0fmWhWAFpk7nCho0yYY5DrOd0KZv8O3D
Fhp2cNinooKdsHrszuCwJQ3p+7wlXXFOOFaz9c03OxrWyg9P5J5zgokr/7EC
n4Y7d+Wcm2qdkBEXdlljJw2DioNZiRQbX50VTs/ZTUPIiYLetbpssLTj/9AK
oiHKd1Fo0Do2ptw+E748goaOvqimVj4bHr3lTpuSaPhlkNZ1KIWN5ZunTXiW
0NDuHxDaqcRB3tXsxc/LaahfRwTepDgwmMFysq6m4fOGAZGNFgcm/6ZeWv6E
hqVvPSvyNnDgcHyZ7VAvDV6B2hev+HDQ9OVRoGyABv6alb5poRzw3QLONn2c
5E1YifxoDtzW3Pl4Y4wGvbiOVevTOdjTZp8aStKhU/vP0Qc1HHzZ8qninQYd
+6hnTpuaOAjPSekXz6VDfuXcpLhODg5HdJuaLqGjMuPwnJwhDlJ0/f+ZYkqH
+icZb7kGF7f9T66O30HHRtPjC0YkXJh2GLmM+9KR/rzliLUnF/fMu6J9A+lI
rY9tj/Llok5jURsrgo4DSzs+Zh7govPurQj1ZDoe9jA51We4kFPpeHi2nI6w
D/2+L1q5sLyg5XVdg4H3xPyhVFMeDFIZi6rnMLAu80C9ljkPCxJ+vGnXYmDf
7vlmGdY8fA9tcVfUY6BHLbwvhMvDNadYVz4YSC4dGczZwQMl/5E3vp0Bclje
xiyRh25publ1MQNt3csU2F08PORdU3AtY4Brlnu7poeHItu0e3sqGXCdPW5s
2MtDomEg0h8xUPh6lKb4gYetqktMPrxi4MNIOymQc0Z26e9rkglVqNtNVMlp
O2PvXKl2j4sqFtay67/KnOHO0nfulKmiyKcvr9LDGQ6H5I+1eqrCoHeb2/Ed
zlj09uLHht2qqJpzV6ru74zmG33lxYdVwXiUnSkf6Qx9uz3SkzmqsBhu+kdw
xhl9YdHplmOq0FkWWejxzBl6Azyi1V4Nww1/v/rLnA9KdWh65Xc1bLlhJ/xk
KEDT3WMx1uNqGGJLN93aIEDiLt2pj3+qwaH2e/mBTQKQdSKFdkUmbDpWVKlb
TuLoqvEPDCbUix9oejgLoPYj7dNsHSYm9vETkoIEoP27tTPAlgndjsSDsYUC
PEx+Kfpmz0T4Bv3ZxcUCxJuHtx1gMRHxZ/nwYKkAxIW8Z7G8Sb9pzSWPGgFU
JLOfZEiZSEu7l3P0uQDTnw5V1QcwsSE3J6HghwCKZadzddIncZ4Ce6aFEO0m
5lbdZ5lQceyzF9kKkVs88DLtPBPxP698ucgSQlS4kSSzmHA8Qy93EAtx/Vp3
oFw+EzO8S92V/IVwTddZ31XDRHDotm7mWSEMNB82pdYxEfb+XWvnRSGUUwN3
chsm/fubg6//JcSNpKr0+iYmBnTeGYbcFIKI8/p5u4uJyOLTzyMahCgKza5M
+czEas8S/bSfQiSMciTsL0z8anSzmDlVhN+Cxkbo35iYuyo/KJkmAsPffnH0
BBNW3Vnr780SwWP7h9gAZRJRN+w5hQYikPw1jqz5JNb75Dqt8xFhp0FJm7It
ibVTj1QbvxOhvLElp8iehMGYReH+TyKo+wxFeLFIGGo4KrZ8F6Hssu6iKh6J
3JaJKWWEGJTGH3si3UjwT814HbtCjJJv24jRIBLJ25ettAsRQy0l4sXlEBKe
PJrt0EExPFanXeeGk5ARnVXX4sRQ3fmYnx9JoqFEmBWcIcZvncaXfOJI6M+r
ENvXikGUMTa+OUdiKdPdJ2q+C9zES9WSLpJYMM3g8ddlLrg5avZ6cyaJvbqN
K0+sc4FUPzQ2I5tEnOGFpGUOLig4/7pFVEjChHZtBxXhAvHhIt8n9SQcgg0y
dvS5IG/B082Rj0ikl/hvE35xgdLd9zNWPiFRHz1AC1GUIHdEuyT+GQkrRyt1
Cy0JFLyOT7PqIXHyZpiqmUSCbGv3c6VfSAQ7aXAauiUw0NMy3fSNBJV3Xzx9
SIJSxsuOsh8kFqkpFOyXc0XjU4lGhRyFCqvEU3O0XTEsESbcV6EA2sB19nZX
7MdMPXM6hTNtfwtiwlwxVaf1QaUqBW7x0PQpx10xq4+rVD2DgtO9K46N+a4w
3cPaVzuPAvv2nFUKU6So4apqWmtRYDU9iE77RQqW4aNbD3QorF0YrRyzXAr3
MdvPdUsoNGu6h293liLmiKV3w2oKO+yCMn/Pk6IxzVTQZEHhW2UYzdvEDcL9
4yNsawp3lBZfGLN2wyvpnaSnthSuth9mPeG7YXjh+sfNLAp+75brWAe6YVbu
WotWEQVJnDl1Pc8NF04Mv+ZLKEws2CtKLHeDXmDBwedSCjfWv8qvbXSDqfGq
u23bJvWDDhJ8coP7fT2DTl8KWd75g4XGMvRf7n/i4k9h99JR09N2MgQeverX
FUBhecrMkiGpDDH2S7JfhFCYe1XZV/6oDDmtOto9URSUDok+27yQYW3xqzJZ
NIX2j7peViMylKafl7yKofDHr3Gdb2juaJTNP/U6nkLQeP3i2abuGOnXVOtN
oWA4cGnz73+5Iz/lWHlzKvW/f7LcHf8B9CpvIA==
       "]], LineBox[CompressedData["
1:eJwtVnk0lQv3pgwlKZz35Zz3lAwXZUhKQpcnJEM453DOMR5jVNccbjJliitC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       "]], LineBox[CompressedData["
1:eJwVk3s4lfkWx91ymyK3GXXe3/uelKQkGQyZ9I3dtm3bvryvaXehvV1Kj3HJ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       "]], 
      LineBox[{{4.951551020408163, 139.37596259923768`}, {4.952725024990185, 
       150.}}], 
      LineBox[{{7.196012175022929, 0.}, {7.19644978787316, 
       0.680122364067862}, {7.198928048521438, 4.3576005303963}, {
       7.201406309169715, 8.366240097856897}, {7.201448979591836, 
       8.448952703871816}}], 
      LineBox[{{2.794950454536869, 0.}, {2.7969872680162746`, 
       52.52496966501432}, {2.7991671838755243`, 150.}}]},
     Annotation[#, "Charting`Private`Tag$13001#1"]& ], {}}, {}},
  AspectRatio->Full,
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{1.317, 0},
  AxesStyle->Directive[
    GrayLevel[0], 
    AbsoluteThickness[0.2]],
  BaseStyle->Automatic,
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox["\"Scattering Length (x BohrRadius)\"", TraditionalForm], None}, {
     FormBox["\"B (G)\"", TraditionalForm], None}},
  FrameStyle->Directive[
    GrayLevel[0], 
    AbsoluteThickness[0.2]],
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  FrameTicksStyle->GrayLevel[0],
  GridLines->{Automatic, Automatic},
  GridLinesStyle->Automatic,
  ImagePadding->All,
  ImageSize->{800, 495},
  LabelStyle->{FontSize -> 14},
  Method->{
   "DefaultBoundaryStyle" -> Automatic, 
    "DefaultGraphicsInteraction" -> {
     "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
      "Effects" -> {
       "Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2}, 
        "Droplines" -> {
         "freeformCursorMode" -> True, 
          "placement" -> {"x" -> "All", "y" -> "None"}}}}, "DefaultMeshStyle" -> 
    AbsolutePointSize[6], "PointSizeFunction" -> None, "ScalingFunctions" -> 
    None, "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Identity[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Identity[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{1.317, 2.173}, {0, 150}},
  PlotRangeClipping->True,
  PlotRangePadding->{{0, 0}, {0, 0}},
  Ticks->{Automatic, Automatic},
  TicksStyle->{FontSize -> 14}]], "Output",
 CellChangeTimes->{{3.9463001299682417`*^9, 3.9463001551784124`*^9}, 
   3.9463004341960926`*^9, {3.9463004674544535`*^9, 3.946300485191087*^9}, 
   3.9463005365743866`*^9, 3.946301604695448*^9, 3.946542264199359*^9},
 CellLabel->"Out[23]=",ExpressionUUID->"61a6f7d5-c877-497e-90e9-8dd81c619ca2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"scatteringLength", "[", "1.317", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"scatteringLength", "[", "2.173", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"BValues", "=", 
    RowBox[{"Range", "[", 
     RowBox[{"1.317", ",", "2.173", ",", "0.001"}], "]"}]}], ";"}], " ", 
  RowBox[{"(*", 
   RowBox[{"Define", " ", "range", " ", "for", " ", "B"}], 
   "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"asValues", "=", 
    RowBox[{"scatteringLength", "/@", "BValues"}]}], ";"}], " ", 
  RowBox[{"(*", 
   RowBox[{
   "Compute", " ", "scattering", " ", "length", " ", "for", " ", "each", " ", 
    "value", " ", "of", " ", "B"}], "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"Transpose", "[", 
    RowBox[{"{", 
     RowBox[{"asValues", ",", "BValues"}], "}"}], "]"}], ",", 
   RowBox[{"FrameLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
     "\"\<Scattering Length (x BohrRadius)\>\"", ",", "\"\<B (G)\>\""}], 
     "}"}]}], ",", 
   RowBox[{"LabelStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"FontSize", "\[Rule]", "14"}], "}"}]}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", " ", 
   RowBox[{"PlotTheme", "\[Rule]", "\"\<Classic\>\""}], ",", 
   RowBox[{"TicksStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"FontSize", "\[Rule]", "14"}], "}"}]}], ",", 
   RowBox[{"PlotStyle", "\[Rule]", 
    RowBox[{"{", "Thick", "}"}]}], ",", 
   RowBox[{"GridLines", "\[Rule]", "Automatic"}], ",", 
   RowBox[{"ImageSize", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"800", ",", "495"}], "}"}]}], ",", 
   RowBox[{"AspectRatio", "\[Rule]", "Full"}]}], "]"}]}], "Input",
 CellChangeTimes->{{3.946300369035057*^9, 3.9463005626365848`*^9}, {
  3.946301228472893*^9, 3.946301357506851*^9}, {3.94630141955521*^9, 
  3.9463014214585342`*^9}, {3.946301914200554*^9, 3.9463019301266484`*^9}},
 CellLabel->"In[24]:=",ExpressionUUID->"329ec9a4-f739-4bf2-bff9-e5d541d4c8fa"],

Cell[BoxData[
 GraphicsBox[{{}, {{}, {}, 
    {RGBColor[0.24720000000000014`, 0.24, 0.6], Thickness[Large], 
     LineBox[CompressedData["
1:eJw12ndcTu8bB/DzGMkKycrIikL5oYzIdSsjSvN5zsjeK1siEUKKhKxkUyoi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      "]]}}, {{}, {}}},
  AspectRatio->Full,
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{67.50782601366832, 1.2694444444444448`},
  AxesStyle->Directive[
    GrayLevel[0], 
    AbsoluteThickness[0.2]],
  BaseStyle->Automatic,
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox["\"B (G)\"", TraditionalForm], None}, {
     FormBox["\"Scattering Length (x BohrRadius)\"", TraditionalForm], None}},
  FrameStyle->Directive[
    GrayLevel[0], 
    AbsoluteThickness[0.2]],
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  FrameTicksStyle->GrayLevel[0],
  GridLines->{Automatic, Automatic},
  GridLinesStyle->Automatic,
  ImageSize->{800, 495},
  LabelStyle->{FontSize -> 14},
  Method->{
   "OptimizePlotMarkers" -> True, "OptimizePlotMarkers" -> True, 
    "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        Identity[
         Part[#, 1]], 
        Identity[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        Identity[
         Part[#, 1]], 
        Identity[
         Part[#, 2]]}& )}},
  PlotRange->{{67.50782601366832, 118.299952122563}, {1.2694444444444448`, 
   2.173}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Ticks->{Automatic, Automatic},
  TicksStyle->{FontSize -> 14}]], "Output",
 CellChangeTimes->{{3.9463013099729557`*^9, 3.9463013581064157`*^9}, 
   3.946301421896168*^9, 3.9463016066577034`*^9, {3.9463019162280746`*^9, 
   3.946301930548483*^9}, 3.946302069709379*^9, 3.946542268549102*^9},
 CellLabel->"Out[28]=",ExpressionUUID->"b41db075-8f6a-413f-b3d4-d969247576da"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"targetas", "=", "71.54"}], ";", 
  RowBox[{"(*", 
   RowBox[{
   "Define", " ", "the", " ", "target", " ", "scattering", " ", "length"}], 
   "*)"}], "\[IndentingNewLine]", 
  RowBox[{"closestIndex", "=", 
   RowBox[{"First", "[", 
    RowBox[{"Ordering", "[", 
     RowBox[{
      RowBox[{"Abs", "[", 
       RowBox[{"asValues", "-", "targetas"}], "]"}], ",", "1"}], "]"}], 
    "]"}]}], ";", " ", 
  RowBox[{"(*", 
   RowBox[{
   "Find", " ", "the", " ", "closest", " ", "value", " ", "in", " ", 
    "a_sValues", " ", "and", " ", "return", " ", "the", " ", "corresponding", 
    " ", "B", " ", "value"}], "*)"}], "\[IndentingNewLine]", 
  RowBox[{"closestB", "=", 
   RowBox[{"BValues", "[", 
    RowBox[{"[", "closestIndex", "]"}], "]"}]}], ";", " ", 
  RowBox[{"(*", 
   RowBox[{"Index", " ", "of", " ", "closest", " ", "a_s", " ", "value"}], 
   "*)"}], 
  RowBox[{"(*", 
   RowBox[{
   "Retrieve", " ", "the", " ", "corresponding", " ", "B", " ", "value"}], 
   "*)"}], "\n", "closestB", " ", 
  RowBox[{"(*", 
   RowBox[{"Print", " ", "the", " ", "closest", " ", "B", " ", "value"}], 
   "*)"}]}]], "Input",
 CellChangeTimes->{{3.9463020828630404`*^9, 3.946302083661174*^9}, {
  3.946302156269303*^9, 3.94630215681246*^9}, {3.9463034223154163`*^9, 
  3.946303436732597*^9}, {3.9463035203877*^9, 3.946303520594941*^9}, {
  3.9465422755424795`*^9, 3.9465422880260324`*^9}, {3.9465428460626583`*^9, 
  3.9465428496456966`*^9}},
 CellLabel->"In[31]:=",ExpressionUUID->"ab4fae22-f7a0-4405-95b8-21ba98484316"],

Cell[BoxData["1.367`"], "Output",
 CellChangeTimes->{{3.946302071185437*^9, 3.9463020838485684`*^9}, 
   3.9463021572281528`*^9, {3.9463034228704424`*^9, 3.9463034369863443`*^9}, 
   3.946303520848451*^9, {3.9465422768162155`*^9, 3.9465422889198856`*^9}, 
   3.946542850289748*^9},
 CellLabel->"Out[31]=",ExpressionUUID->"538f6375-f1c3-4276-85ef-0fef1f723b6b"]
}, Open  ]]
},
WindowSize->{1904, 981},
WindowMargins->{{1920, Automatic}, {Automatic, 0}},
TaggingRules->{
 "WelcomeScreenSettings" -> {"FEStarting" -> False}, "TryRealOnly" -> False},
FrontEndVersion->"12.2 for Microsoft Windows (64-bit) (December 12, 2020)",
StyleDefinitions->"Default.nb",
ExpressionUUID->"845021d5-1056-4669-8b9e-6ad0bd0c74a1"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[558, 20, 2766, 86, 395, "Input",ExpressionUUID->"62f7b3c0-151e-42a7-bd14-2c7024749c6f"],
Cell[CellGroupData[{
Cell[3349, 110, 4164, 99, 143, "Input",ExpressionUUID->"461e8a7b-9e39-408e-a5f3-38960ab44ff7"],
Cell[7516, 211, 24540, 433, 512, "Output",ExpressionUUID->"fe3d2ab1-2e60-4c87-b304-9af09b512293"]
}, Open  ]],
Cell[CellGroupData[{
Cell[32093, 649, 1572, 37, 48, "Input",ExpressionUUID->"e3efdc46-0d6e-4d16-ae45-a127d2ef7398"],
Cell[33668, 688, 24453, 431, 512, "Output",ExpressionUUID->"61a6f7d5-c877-497e-90e9-8dd81c619ca2"]
}, Open  ]],
Cell[CellGroupData[{
Cell[58158, 1124, 2017, 51, 124, "Input",ExpressionUUID->"329ec9a4-f739-4bf2-bff9-e5d541d4c8fa"],
Cell[60178, 1177, 14370, 257, 512, "Output",ExpressionUUID->"b41db075-8f6a-413f-b3d4-d969247576da"]
}, Open  ]],
Cell[CellGroupData[{
Cell[74585, 1439, 1546, 37, 105, "Input",ExpressionUUID->"ab4fae22-f7a0-4405-95b8-21ba98484316"],
Cell[76134, 1478, 360, 5, 32, "Output",ExpressionUUID->"538f6375-f1c3-4276-85ef-0fef1f723b6b"]
}, Open  ]]
}
]
*)

